SystemOrganization addCategory: #Namespaces!

!ClassOrganizer methodsFor: '*namespaces' stamp: 'mvdg 6/6/2008 21:26'!
copy
	" Copy. You need to set the subject yourself. "
	| newInstance |
	newInstance := self class new.
	newInstance setSubject: self subject.
	newInstance classComment: classComment stamp: commentStamp.
	newInstance setSubject: subject. " You may need to change this yourself. "
	newInstance 
		privCatArray: categoryArray copy
		catStops: categoryStops copy
		elementArray: elementArray copy.
	^ newInstance
		! !

!ClassOrganizer methodsFor: '*namespaces' stamp: 'mvdg 6/6/2008 21:18'!
privCatArray: catArray catStops: catStops elementArray: eArray
	categoryArray := catArray.
	categoryStops := catStops.
	elementArray := eArray.! !

IdentityDictionary subclass: #Namespace
	instanceVariableNames: 'name imports package fullyQuantifiedName parent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Namespaces'!

!Namespace commentStamp: 'mvdg 6/4/2008 21:06' prior: 0!
I implemented nested Namespaces.

I contain a dictionary of #names to either Classes or shared variables. Nesting is done by following #names that map to other Namespaces.

"imports" is a list of imports of other namespaces. These will be searched when my own classes try to resolve a reference.

For example:

myDemoMethod
	^ Speech.Events.VoiceEvent new.
	
In this example, the root namespace (which should be included in the import list by default) will be searched to find #'Speech.Events.VoiceEvent', which should return a class.

If you include the "Speech.Events" Namespace in the import list, then you can do:

myDemoMethod
	^ VoiceEvent new.
	
Also, the "current" Namespace is also searched, so if VoiceEvent was also in the current Namespace, then that would be returned. This also works for nesting - if VoiceEvent had sub-namespaces, then you could refer to VoiceEvents.Demos.DemoEvent (for example).
	
Namespaces are intended to replace the #Smalltalk SystemDictionary. Because every class has access to #Smalltalk, it is near impossible to implement capability-based security. Instead, a class will only have access to it's enveloping Namespace which is used to look up class names and global variables.

Namespaces also implement replication for classes (when using DPON). Because classes are special objects, their replication algorithm is hard-wired to CopyRepAlg. Instead, classes need to be updated by their enclosing namespaces when code is remotely changed. There can be multiple implementations of Namespaces with different replication algorithms, provided that they implement the same type.

Shared variables
-------------------
It is now possible to use shared variables. To use them:
1. Open up the Namespace in an inspector, scoped workspace or whatever.
2. Add the shared variable (My.Namespace at: #VarName put: nil).
3. Write a class-side initialize method to set the shared variable values.

You can currently read from any shared variable. However, methods can only write to shared variables which are referred to directly in the same namespace or a sub-namespace of that method. 

For example, if you define a class in MyNamespace.MyClass, you can refer to the shared variables in MyNamespace.MySharedVar (in your code: 'MySharedVar := ...'), or MyNamespace.Private.MySharedVar (in your code: 'Private.MySharedVar := ...'). In this example, you can write but not read shared variables elsewhere, such as in parent namespaces (the root namespace in this example) or other Namespaces on the import list.

Technically, this is because shared variables you can write to are linked in your CompiledMethods using an Association, while read-only shared variables are linked using ReadOnlyVariableBindings.

The class-side initialize method will be called every time you relink, recompile or copy a Package.

Filing our and filing in should work if you have written your class-side initialize routine.!

!Namespace class methodsFor: 'as yet unclassified' stamp: 'mvdg 10/8/2007 20:32'!
replicationAlgorithm
	^ NamespaceRepAlg.! !

!Namespace class methodsFor: 'as yet unclassified' stamp: 'mvdg 9/23/2007 14:59'!
separator
	" Return the string used to separate elements in a namespace name "
	^ '.'! !

!Namespace methodsFor: 'comparing' stamp: 'mvdg 12/30/2007 18:54'!
<= other
	^ self name <= other name.! !

!Namespace methodsFor: 'comparing' stamp: 'mvdg 9/3/2007 22:09'!
= anotherNamespace
	(anotherNamespace isKindOf: Namespace) ifFalse: [ ^ false ].
	^ (self package equalTo: anotherNamespace package)
	and: [ self fullyQuantifiedName = anotherNamespace fullyQuantifiedName ].! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 5/6/2007 18:53'!
acceptVisitor: aVisitor
	^ aVisitor visitNamespace: self.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 4/27/2008 21:52'!
addClass: aClass
	" Don't forget to relink to make sure the superclass name and literals are valid. "
	aClass namespace: self.
	^ self at: (aClass name asSymbol) put: aClass.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 10/26/2007 23:30'!
addImport: anImport
	"anImport value. Hack: make it not a Future.
	self assert: [anImport isKindOf: Namespace]."
	" Need an OrderedSet!! "
	(imports includes: anImport) ifFalse: [ imports addLast: anImport ].! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/28/2008 16:09'!
addNamespace: aNamespace
	" self changed. - TODO: should do this, but this will break imports. "
	| p |
	aNamespace parent: self.
	p := self package. 
	p ifNotNil: [ " Then this namespace is temporarily dislocated from a package,
		such as when it is being copied. "
		aNamespace recurse: [ :each | each package: p ] ].
	^ self at: (aNamespace name asSymbol) put: aNamespace.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 12:38'!
bindingOf: varName 
	| notFound rv |
	notFound := false.
	rv := self
				lookupAssociation: varName
				ifAbsent: [nil].
	rv ifNil: [
		" We are searching the imports; a Package's associations are only 
		  modifyable by classes in that same package (i.e. 'private'), so we
		  use a ReadOnlyVariableBinding here: "
		rv := ReadOnlyVariableBinding key: varName value: 
			(self
				lookupInImports: varName
				ifAbsent: [^ nil])].
	self assert: [ (rv isMemberOf: Association) or: [rv isMemberOf: ReadOnlyVariableBinding]].
	^ rv.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 2/6/2007 21:22'!
childClasses
	^ self values
		select: [:each | each isKindOf: Behavior]! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 12/30/2007 11:54'!
childNamespaces
	^ self values
		select: [:each | each isKindOf: Namespace].! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 20:54'!
childSharedVariables
	^ self values
		select: [:each | ((each isKindOf: Behavior) not) and: [(each isKindOf: Namespace) not] ]! !

!Namespace methodsFor: 'private' stamp: 'mvdg 12/25/2007 17:58'!
collectionAsDottedNotation: aCollection
	" This is a utility function to convert the given collection of namespaces into a single string, separated by dots. "
	| resultStream |
	aCollection isEmpty ifTrue: [ ^ '' ].
	resultStream := WriteStream on: String new.
	aCollection do: [ :each |
		resultStream nextPutAll: each name asString.
		resultStream nextPut: $.
	].
	^ resultStream contents.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 4/5/2008 20:30'!
copy
	| newInstance |
	newInstance := self class new.
	newInstance copyFrom: self.
	^ newInstance.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/30/2008 20:06'!
copyFrom: anotherNamespace
	name := anotherNamespace name.
	parent ifNotNil: [package := parent package].
	
	anotherNamespace associationsDo: [ :each |
		(each value isMemberOf: Namespace) ifTrue: [
			| newNamespace |
			newNamespace := Namespace new.
			newNamespace name: each value name. " To make sure the next line works:"
			self addNamespace: newNamespace.
			newNamespace copyFrom: each value."recursive" ]
		ifFalse: [ (each value isKindOf: Behavior) ifTrue: [
			| newClass |
			newClass := each value deepCopy.
			self addClass: newClass.
			newClass superclassName: each value superclassName ]
		ifFalse: [
			self add: each ]]].
	
	" Do not copy the imports here. Any imports referring elsewhere in this package need to be updated, and this method does not know if the whole package or only part of a package is being copied (and whose referrers thus need updating). Use >>copyImports:From: later if necessary. "! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 4/5/2008 20:40'!
copyImportsFrom: anotherNamespace
	anotherNamespace imports do: [ :eachImport | 
		| path updatedImport |
		(anotherNamespace package = eachImport package) ifTrue: [
			path := eachImport fullyQuantifiedName.
			updatedImport := package lookup: path
				ifAbsent: [ self error: 'Could not update imports in this Namespace.' ].
			self addImport: updatedImport.
		] ifFalse: [
			self addImport: eachImport. ] ].

	! !

!Namespace methodsFor: 'hacks' stamp: 'mvdg 9/2/2007 20:54'!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system.
	Note that this doesn't do everything required to dispose of a class - to do that use Class>>removeFromSystem.
	This is based on the implementation in SystemDictionary and is called from the above method on Class."

	self removeKey: aClass name ifAbsent: [].! !

!Namespace methodsFor: 'initialization' stamp: 'mvdg 4/25/2007 14:15'!
fullyQuantifiedName
	fullyQuantifiedName isNil ifTrue: [ ^ name copy ].
	^ fullyQuantifiedName copy.! !

!Namespace methodsFor: 'initialization' stamp: 'mvdg 3/15/2008 15:56'!
fullyQuantifiedName: aName
	self assert: [aName isKindOf: String].
	fullyQuantifiedName := aName.! !

!Namespace methodsFor: 'comparing' stamp: 'mvdg 6/14/2007 02:08'!
hash
	^ self package uuid hash bitXor: self fullyQuantifiedName hash. ! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 3/18/2007 22:00'!
imports
	^ imports copy.! !

!Namespace methodsFor: 'initialization' stamp: 'mvdg 3/24/2007 17:01'!
initialize: aNumber
	imports := OrderedCollection new.
	^ super initialize: aNumber.! !

!Namespace methodsFor: 'testing' stamp: 'mvdg 11/4/2008 19:47'!
invariant
	" Make sure everything is sane. "
	self assert: [ (self package lookup: self fullyQuantifiedName) = self ].
	self childClasses do: [ :each |
		self assert: [ each namespace = self ].
		self assert: [ each class namespace = self ].
		self assert: [ each class superclass = each superclass class ] ].
	self childNamespaces do: [ :each |
		self assert: [each parent = self].
		self assert: [each package = self package].
		each invariant].
	! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 11:59'!
lookup: aName
	^ (self lookup: aName ifAbsent: [self error: 'Could not find ', aName asString]) value.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 11:59'!
lookup: dottedName ifAbsent: aBlock
	" Look up that fully quantified name in self. Only childNamespaces are searched; imports are not considered here. The name starts searching from my childNamespaces. "
	dottedName isEmpty ifTrue: [ ^ self ]. " Return the root "
	^ (self lookupFromNamesArray: (dottedName subStrings: '.') level: 1 ifAbsent: aBlock) value.
! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 12:00'!
lookup: dottedName ifPresent: aBlock
	" Look up that fully quantified name. The look-up does not include self; it starts with self's children. "
	| result |
	result := self lookupFromNamesArray: (dottedName subStrings: '.') level: 1 ifAbsent: [nil].
	result isNil ifFalse: [ ^ aBlock value: result value ].
	^ nil.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 12:20'!
lookupAssociation: aName
	^ self lookupAssociation: aName ifAbsent: [self error: 'Could not find ', aName asString].! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 12:00'!
lookupAssociation: dottedName ifAbsent: aBlock
	" Look up that fully quantified name in self. Only childNamespaces are searched; imports are not considered here. The name starts searching from my childNamespaces. "
	dottedName isEmpty ifTrue: [ ^ #'' -> self ]. " Return the root "
	^ self lookupFromNamesArray: (dottedName subStrings: '.') level: 1 ifAbsent: aBlock.
! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/2/2008 12:02'!
lookupAssociation: dottedName ifPresent: aBlock
	" Look up that fully quantified name. The look-up does not include self; it starts with self's children. "
	| result |
	result := self lookupFromNamesArray: (dottedName subStrings: '.') level: 1 ifAbsent: [nil].
	result isNil ifFalse: [ ^ aBlock value: result ].
	^ nil.! !

!Namespace methodsFor: 'private' stamp: 'mvdg 6/2/2008 11:58'!
lookupFromNamesArray: fqnArray level: lvl ifAbsent: aBlock
	" Private. Look up that name using a recursive algorithm "
	| next |
	lvl = fqnArray size ifTrue: [ ^ self associationAt: (fqnArray at: lvl) asSymbol ifAbsent: aBlock ].
	next := self at: (fqnArray at: lvl) asSymbol ifAbsent: [ ^ aBlock value ].
	(next isKindOf: Namespace) 
		ifTrue: [ ^ next lookupFromNamesArray: fqnArray level: (lvl + 1) ifAbsent: aBlock. ]
		ifFalse: [ ^ aBlock value. ].! !

!Namespace methodsFor: 'private' stamp: 'mvdg 3/18/2007 22:09'!
lookupInImports: fqn ifAbsent: aBlock
	| notFound rv |
	imports do: [ :each | 
		notFound := false.
		rv := each lookup: fqn ifAbsent: [ notFound := true ].
		notFound ifFalse: [ ^ rv ].
	].
	^ aBlock value.! !

!Namespace methodsFor: 'initialization' stamp: 'mvdg 3/23/2007 21:23'!
name
	^ name.! !

!Namespace methodsFor: 'initialization' stamp: 'mvdg 5/9/2007 21:06'!
name: aName
	name := aName asSymbol.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 12/31/2007 15:38'!
nameRelativeTo: anotherNamespace
	" Return the dotted notation name that can be used from anotherNamespace to access me. This method is slow because it searches all children and imports. "
	anotherNamespace recurse: [ :each :path |
		(self == each) ifTrue: [
			^ self collectionAsDottedNotation: path ] 
	] path: OrderedCollection new.

	anotherNamespace imports do: [ :each |
		each recurse: [ :eachImport :path |
			(self = eachImport) ifTrue: [
				^ self collectionAsDottedNotation: path ]
		] path: OrderedCollection new.
	].
	^ nil.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 4/5/2008 15:58'!
package
	^ package.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 6/15/2008 15:30'!
package: aPackage
	self assert: [aPackage isKindOf: Package].
	package := aPackage.! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 3/24/2007 16:42'!
parent
	^ parent! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 3/15/2008 21:30'!
parent: aNamespace
	| |
	self assert: [aNamespace isKindOf: Namespace].
	parent := aNamespace.
	
	" Also set my FQN!! "
	(parent isNil or: [parent fullyQuantifiedName isEmpty]) ifTrue: [
		fullyQuantifiedName := name.
	] ifFalse: [
		fullyQuantifiedName := parent fullyQuantifiedName, '.', name
	].! !

!Namespace methodsFor: 'printing' stamp: 'mvdg 4/28/2008 20:32'!
printOn: aStream
	fullyQuantifiedName isNil 
		ifTrue: [ aStream nextPutAll: name asString ]
		ifFalse: [ aStream nextPutAll: (fullyQuantifiedName asString) ].
	aStream
		nextPut: $<;
		nextPutAll: self asOop asString;
		nextPut: $>! !

!Namespace methodsFor: 'enumerating' stamp: 'mvdg 11/24/2007 16:24'!
recurse: aBlock
	aBlock value: self.
	self childNamespaces do: [ :each | each recurse: aBlock].! !

!Namespace methodsFor: 'enumerating' stamp: 'mvdg 12/25/2007 17:32'!
recurse: aBlock path: aPath
	aBlock value: self value: aPath.
	self childNamespaces do: [ :each | each recurse: aBlock path: (aPath copyWith: each)].! !

!Namespace methodsFor: 'enumerating' stamp: 'mvdg 6/6/2008 21:32'!
recurseTogether: anotherNamespace doing: aBlock ifError: errorBlock
	| myChildren herChildren |
	" Recurse through this namespace and all children, and anotherNamespace's children, in lock-step. "
	aBlock value: self value: anotherNamespace.
	myChildren := self childNamespaces sortBy: [ :a :b | a name < b name ].
	herChildren := anotherNamespace childNamespaces sortBy: [ :a :b | a name < b name ].
	(myChildren size ~= herChildren size) ifTrue: [ ^ errorBlock value ].
	
	1 to: myChildren size do: [ :each |
		| myChild herChild |
		myChild := myChildren at: each.
		herChild := herChildren at: each.
		(myChild name ~= herChild name) ifTrue: [ ^ errorBlock value ].
		myChild recurseTogether: herChild doing: aBlock ifError: errorBlock ].! !

!Namespace methodsFor: 'accessing' stamp: 'mvdg 12/23/2007 16:49'!
removeImport: anImport
	imports remove: anImport.! !

Namespace subclass: #Package
	instanceVariableNames: 'uuid deserializeSemaphore readOnly'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Namespaces'!

!Package commentStamp: 'mvdg 2/23/2008 11:32' prior: 0!
I implement a single Package of code. I form a root Namespace which other namespaces can add to their import lists if they want to use classes inside me.

The instvars author, date, version, license, history and locallyModified may all be removed. This information is all stored in my associated PackageVersion.

A package is first instantiated as writable. Once it has been made read-only, it cannot be made writable again. If you want to make the package writable again then you need to make a copy.!

!Package methodsFor: 'accessing' stamp: 'mvdg 5/6/2007 21:08'!
acceptVisitor: aVisitor
	^ aVisitor writePackage: self.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 2/13/2008 21:35'!
changed
	readOnly ifTrue: [ self error: 'This package is read-only.' ].
	^ super changed.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 5/12/2008 21:05'!
copyFrom: anotherPackage
	" Set all of my instance variables to be the same as anotherObject. "
	readOnly := false.
	package := self.
	super copyFrom: anotherPackage.
	self recurseTogether: anotherPackage doing: [ :me :her | me copyImportsFrom: her ]
		ifError: [ self error: 'Could not transfer imports between packages.' ].
	self relinkNotifying: Complainer new.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 3/16/2008 16:20'!
description
	" Return a String describing this package in more detail "
	| s |
	s := WriteStream on: String new. 
	s	nextPutAll:  self name;
		nextPutAll: ' package'.
	self readOnly ifTrue: [ s nextPutAll: ' (read only)' ].
	s nextPut: $..
	^ s contents.
	" TODO: ^ self name, '-', self author asString, ' v', self version asString."! !

!Package methodsFor: 'comparing' stamp: 'mvdg 4/5/2008 20:09'!
equalTo: anotherPackage 
	" Assume anotherPackage is a Package. >>= is implemented in my superclass. "
	anotherPackage ifNil: [ ^ false ].
	^ self uuid = anotherPackage uuid.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 5/19/2007 22:22'!
fullyQuantifiedName
	^ ''. "I am root"! !

!Package methodsFor: 'initialization' stamp: 'mvdg 2/23/2008 21:02'!
initialize: aNumber
	uuid := UUID new.
	readOnly := false.
	^ super initialize: aNumber.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 6/3/2008 20:56'!
initializeEachClass
	" Call #initialize on every class if it is implemented. "
	self recurse: [ :each |
		each childClasses do: [ :eachClass |
			((eachClass isKindOf: Class)
				and: [ eachClass canUnderstand: #initialize ])
				ifTrue: [ eachClass initialize ] ] ].! !

!Package methodsFor: 'accessing' stamp: 'mvdg 6/9/2007 14:11'!
package
	^ self.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 3/7/2008 21:57'!
readOnly
	readOnly ifNil: [ ^false ].
	^ readOnly.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 2/23/2008 21:12'!
readOnly: aBool
	self assert: [ aBool isKindOf: Boolean ].
	(readOnly & aBool not) ifTrue: [
		" Packages cannot be made writable again after they have been made read-only. If you want to modify this package, make a new version. "
		self error: 'This package is read-only; cannot make writable.' ].
	readOnly := aBool.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 6/16/2008 20:30'!
relinkNotifying: ui
	" Ask every method under me to relink its literals. "
	self invariant.
	self recurse: [ :eachNamespace |
		eachNamespace childClasses do: [ :eachClass |
			eachClass relinkNotifying: ui. ] ].
	" TODO self initializeEachClass."
	self invariant.! !

!Package methodsFor: 'accessing' stamp: 'mvdg 4/28/2007 16:18'!
uuid
	^ uuid! !

!Package methodsFor: 'accessing' stamp: 'mvdg 5/9/2007 21:47'!
uuid: aUUID
	uuid := aUUID.! !

Namespace subclass: #SystemDictionaryNamespace
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Namespaces'!

SystemDictionaryNamespace class
	instanceVariableNames: 'legacyPackage'!

!SystemDictionaryNamespace commentStamp: 'mvdg 5/22/2007 09:58' prior: 0!
I allow classes to use imports from the legacy Smalltalk SystemDictionary.!

SystemDictionaryNamespace class
	instanceVariableNames: 'legacyPackage'!

!SystemDictionaryNamespace class methodsFor: 'as yet unclassified' stamp: 'mvdg 10/18/2007 11:02'!
legacyPackage 
	| p n |
	legacyPackage ifNil: [
		p := Package new name: #Legacy.
		n := SystemDictionaryNamespace new.
		p addNamespace: n.
		n package: p.
		" We need a constant UUID for importing. "
		p uuid: (UUID fromString: 'dc536b66-358e-46f6-8243-2e5d8d810dbd').
		legacyPackage := p.
	].
	^ legacyPackage.! !

!SystemDictionaryNamespace class methodsFor: 'as yet unclassified' stamp: 'mvdg 2/24/2008 14:31'!
legacyPackageRep
	| pv r |
	pv := self legacyPackageVersion.
	r := PackageRepository new.
	r addLast: pv.
	^ r.! !

!SystemDictionaryNamespace class methodsFor: 'as yet unclassified' stamp: 'mvdg 2/24/2008 14:28'!
legacyPackageVersion
	| p pv |
	p := self legacyPackage.
	pv := PackageVersion new.
	pv package: p.
	^ pv.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 5/22/2007 09:59'!
addImport: anImport
	self shouldNotImplement.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 6/4/2007 08:17'!
at: aSymbol
	^ Smalltalk at: aSymbol.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 6/4/2007 08:17'!
at: aSymbol ifAbsent: aBlock
	^ Smalltalk at: aSymbol ifAbsent: aBlock.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 5/22/2007 10:00'!
at: aSymbol put: avalue
	^ self shouldNotImplement.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 5/22/2007 10:02'!
childClasses
	^ Smalltalk values.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 5/22/2007 10:02'!
childNamespaces
	^ #().! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 2/29/2008 22:07'!
copy
	self inform: 'Warning: don''t edit anything!!'.
	^ self.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 10/22/2007 15:47'!
includesKey: aKey
	^ Smalltalk includesKey: aKey. 
! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 6/9/2007 14:10'!
initialize: num
	self name: #Smalltalk.
	self fullyQuantifiedName: 'Legacy.Smalltalk'.
	^ super initialize: num.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 12/31/2007 16:05'!
lookup: aName ifAbsent: aBlock
	^ Smalltalk at: aName asSymbol ifAbsent: aBlock.! !

!SystemDictionaryNamespace methodsFor: 'as yet unclassified' stamp: 'mvdg 5/23/2007 07:47'!
lookupFromNamesArray: namesArray level: n ifAbsent: aBlock
	(namesArray size = n)  ifTrue: [
		^ Smalltalk at: (namesArray at: n) asSymbol ifAbsent: aBlock.
	].
	^ aBlock value.! !

!Metaclass methodsFor: '*namespaces' stamp: 'mvdg 7/3/2008 19:48'!
deepCopy 
	| newMeta |
	newMeta := self shallowCopy.
	newMeta methodDictionary: (self methodDictionary deepCopy).
	newMeta organization: (self organization copy).
	newMeta organization setSubject: newMeta.
	^ newMeta
! !

!Metaclass methodsFor: '*namespaces' stamp: 'mvdg 6/26/2007 21:45'!
namespace
	^ self theNonMetaClass namespace.! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 6/24/2008 21:17'!
<= anotherClass
	^ self name <= anotherClass name.! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 1/9/2008 22:05'!
basicCategory
	(self namespace isMemberOf: SystemDictionaryNamespace) ifFalse: [
		self halt: 'Stop now before you corrupt your class!! Class instvar category has been hijacked for namespaces (to store the superclass name).' ].
	^category! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 1/9/2008 22:20'!
basicCategory: aSymbol
	(self namespace isMemberOf: SystemDictionaryNamespace) ifFalse: [
		self halt: 'Do not continue!! You''ll break the superclass name (for which the category instvar has been hijacked by namespaces)' ].
	category := aSymbol! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 5/18/2008 13:06'!
binding

	^ (self environment associationAt: name ifAbsent: [nil -> self]) copy.
! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 6/5/2008 20:19'!
bindingOf: varName
	"Answer the binding of some variable resolved in the scope of the receiver. This has been modified for Namespaces and attempts to retain the original search order."
	| aSymbol binding |
	aSymbol := varName asSymbol.
	
	"See if it is a class variable in my class pool. "
	binding := self classPool bindingOf: aSymbol.
	binding ifNotNil: [^binding].

	" Look in either the local namespace and import list, or Smalltalk"
	binding := self environment bindingOf: aSymbol.
	binding ifNotNil:[^binding " should already be a copy. "].

	" See if it a class variable in a superclass. "
	superclass isNil ifTrue: [ ^ nil ].
	binding := superclass searchClassPoolForBinding: aSymbol.
	^ binding copy.
! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 1/9/2008 22:28'!
category
	"Answer the system organization category for the receiver. First check whether the
	category name stored in the ivar is still correct and only if this fails look it up
	(latter is much more expensive)"

	| result |
	(self namespace isMemberOf: SystemDictionaryNamespace) ifFalse: [ ^ '' ].
	
	
	" Do not run the following if this class uses namespaces!!!! category has been hijacked to store the superclass name for relinking. "
	
	self basicCategory ifNotNilDo: [ :symbol |
		((SystemOrganization listAtCategoryNamed: symbol) includes: self name)
			ifTrue: [ ^symbol ] ].
	self basicCategory: (result := SystemOrganization categoryOfElement: self name).
	^result! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 1/9/2008 22:01'!
category: aString 
	"Categorize the receiver under the system category, aString, removing it from 
	any previous categorization."

	| oldCategory |
	" category has been hijacked by namespaces for storing the superclass name. "
	self assert: [self namespace isMemberOf: SystemDictionaryNamespace].
	oldCategory := self basicCategory.
	aString isString
		ifTrue: [
			self basicCategory: aString asSymbol.
			SystemOrganization classify: self name under: self basicCategory ]
		ifFalse: [self errorCategoryName].
	SystemChangeNotifier uniqueInstance
		class: self recategorizedFrom: oldCategory to: self basicCategory! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 7/3/2008 19:47'!
deepCopy 
	| newOrg newClass |
	newOrg := self organization copy.
	newClass := self class deepCopy new
		superclass: superclass
		methodDict: self methodDict deepCopy
		format: format
		name: name
		organization: newOrg
		instVarNames: instanceVariables copy
		classPool: classPool copy
		sharedPools: sharedPools.
	newOrg setSubject: newClass.
	^ newClass! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 3/16/2008 20:25'!
fullyQuantifiedName
	self namespace isNil ifTrue: [ ^ self name asString ].
	self namespace fullyQuantifiedName isEmpty 
		ifTrue: [ ^ self name asString ].
	^ self namespace fullyQuantifiedName, '.', self name.! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 3/31/2008 20:44'!
namespace
	(environment isKindOf: Namespace) ifTrue: [
		^ environment
	] ifFalse: [ ^ SystemDictionaryNamespace legacyPackage lookup: #Smalltalk ].! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 5/18/2008 11:58'!
namespace: aNamespace
	self assert: [aNamespace isKindOf: Namespace].
	self assert: [ self ~= (Smalltalk at: self name ifAbsent: [nil]) ].
	environment := aNamespace.! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 11/4/2008 19:56'!
privateRelinkSuperclassNotifying: ui
	" Relink the superclass. "
	| theSuperclass |
	(self namespace isKindOf: SystemDictionaryNamespace)
		ifTrue: [ self error: 'About to re-assign a superclass; you probably don''t want this.' ].
	(self superclassName = 'nil') ifTrue: [ ^ self " superclass is nil; no action. " ].
	
	" I'm not going to call 'self superclass:' here because I don't want to flush the cache yet. "
	theSuperclass := (self namespace bindingOf: (self superclassName)) value.
	theSuperclass ifNil: [ 
		(ui addImportFor: (self superclassName) inNamespace: self namespace)
			ifNil: [ self error: 'Some superclasses in this package could not be correctly relinked.' ].
		theSuperclass := (self namespace bindingOf: (self superclassName)) value.
		theSuperclass ifNil: [ ^ self error: 'Failed trying to relink superclass.' ].	
	].
	superclass := theSuperclass.
	self class privateSetSuperclass: theSuperclass class.! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 5/11/2008 14:29'!
recompileFrom: oldClass
	" Don't invoke compileAllFrom: because it also recompiles my class. "
	oldClass selectorsDo: [:sel | self recompile: sel from: oldClass].! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 9/5/2007 22:32'!
searchClassPoolForBinding: aSymbol
	| binding |
	binding := self classPool bindingOf: aSymbol.
	binding ifNotNil: [^binding].
	superclass isNil
		ifTrue: [ ^ nil ]
		ifFalse: [ ^ superclass searchClassPoolForBinding: aSymbol ].! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 6/15/2008 11:51'!
superclassName
	" Return the dotted-notation name of the superclass (a String) that can be searched in my namespace for that class. This is used for relinking later. "
	" TODO: until I refactor the class hierarchy to use namespaces, I'll stick the superclass name in the category instvar, which is not used in namespaced classes. "
	superclass ifNil: [ ^ nil asString ].
	
	(SystemDictionaryNamespace isMemberOf: SystemDictionaryNamespace) 
		ifTrue: [ ^ superclass name ].
	
	category ifNil: [category := ((superclass namespace nameRelativeTo: self namespace), superclass name) ].
	
	self assert: [((self namespace bindingOf: category) value isKindOf: Class)
		or: [category = 'nil'] ].
	^ category.! !

!Class methodsFor: '*namespaces' stamp: 'mvdg 5/18/2008 11:25'!
superclassName: aFQN
	" Set the name of the superclass to the given fully quantified name. This is used for relinking later. "
	" TODO: until I refactor the class hierarchy to use namespaces, I'll stick the superclass name in the category instvar: "
	self assert: [ (self namespace isMemberOf: SystemDictionaryNamespace) not. ].
	category := aFQN.! !

!CompiledMethod methodsFor: '*namespaces' stamp: 'mvdg 6/24/2008 21:21'!
<= anotherObject
	^ self selector <= anotherObject selector.! !

!CompiledMethod methodsFor: '*namespaces' stamp: 'mvdg 5/3/2008 17:02'!
copy
	" Not implemented in Squeak, so I need to implement one. "
	| newInstance |
	newInstance := self class newMethod: (self numBytecodes) header: (self header).
	newInstance replaceFrom: 1 to: (self size) with: self.
	^ newInstance.
	! !

!CompiledMethod methodsFor: '*namespaces' stamp: 'mvdg 7/9/2008 21:29'!
numBytecodes
	^ (self size) - (self numLiterals + 1 * 4 ).! !

!CompiledMethod methodsFor: '*namespaces' stamp: 'mvdg 6/4/2008 20:45'!
privateFindBindingOf: someLiteral inContext: aContext notifying: ui
	" Find the binding of that literal, potentially asking the user about it. "
	| newLiteralValue |
	newLiteralValue := aContext bindingOf: someLiteral.
	
	^ newLiteralValue ifNil: [
		(ui addImportFor: someLiteral inNamespace: aContext namespace)
			ifNil: [ self error: 'Some literals in this package may not be correctly linked.' ].
		aContext bindingOf: someLiteral.
	].! !

!CompiledMethod methodsFor: '*namespaces' stamp: 'mvdg 5/18/2008 15:01'!
relinkInContext: aContext notifying: ui
	| literals |
	literals := self literals.
	1 to: literals size do: [ :i |
		| eachLiteral |
		eachLiteral := literals at: i.
		((eachLiteral isMemberOf: Association) or: [eachLiteral isMemberOf: ReadOnlyVariableBinding])
			ifTrue: [
				" Then rebind it. "
				" TODO: why do Metaclass instances have nil->class as their last literal? "
				(eachLiteral key isNil) ifFalse: [
					self literalAt: i put: (self privateFindBindingOf: eachLiteral key inContext: aContext notifying: ui). ]
				]].! !

!Behavior methodsFor: '*Namespaces' stamp: 'mvdg 3/28/2008 16:17'!
privateRelinkSuperclassNotifying: ui
	^ self subclassResponsibility.! !

!Behavior methodsFor: '*Namespaces' stamp: 'mvdg 11/4/2008 19:56'!
privateSetSuperclass: aSuperclass
	(self namespace isKindOf: SystemDictionaryNamespace)
		ifTrue: [ self error: 'About to re-assign a superclass; you probably don''t want this.' ].
	superclass := aSuperclass.! !

!Behavior methodsFor: '*Namespaces' stamp: 'mvdg 5/18/2008 15:00'!
relinkNotifying: ui
	" Ask every method under me to relink its literals. "
	" TODO: also relink my superclass. "

	(self namespace isKindOf: SystemDictionaryNamespace)
		ifTrue: [ self error: 'About to seriously break your image.' ].
	self privateRelinkSuperclassNotifying: ui.
	self methodsDo: [ :eachMethod | eachMethod relinkInContext: self notifying: ui].
	
	(self class namespace isKindOf: SystemDictionaryNamespace)
		ifTrue: [ self error: 'About to break your image.' ].
	self class methodsDo: [ :eachMethod | eachMethod relinkInContext: self notifying: ui].
	self flushCache. " I think...? "! !

Object subclass: #Complainer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Namespaces'!

!Complainer commentStamp: 'mvdg 5/12/2008 21:02' prior: 0!
A Complainer is a class used to pass to the compiler as the ui to do callbacks on if something goes wrong. It just throws an error if anything is called back on.!

!Complainer methodsFor: 'as yet unclassified' stamp: 'mvdg 5/12/2008 21:05'!
addImportFor: aSymbol inNamespace: ns
	self error: 'Could not find ', aSymbol, ' in context: ', ns fullyQuantifiedName.! !

!Complainer methodsFor: 'as yet unclassified' stamp: 'mvdg 5/12/2008 21:03'!
notify: aString at: anInteger in: aStream
	self error: aString.! !

!Complainer methodsFor: 'as yet unclassified' stamp: 'mvdg 5/12/2008 21:03'!
select! !

!Complainer methodsFor: 'as yet unclassified' stamp: 'mvdg 5/12/2008 21:03'!
selectFrom: a to: b! !

!Complainer methodsFor: 'as yet unclassified' stamp: 'mvdg 5/12/2008 21:03'!
selectionInterval
	^ 1 to: 1.! !

!MethodDictionary methodsFor: '*namespaces' stamp: 'mvdg 5/3/2008 17:17'!
deepCopy
	" Make a copy of me, with new copies of all my CompiledMethods too. "
	| newInstance |
	newInstance := self class new.
	self keysAndValuesDo: [ :key :value |
		newInstance at: key put: (value copy) ].
	^ newInstance.! !

!Scanner methodsFor: '*namespaces' stamp: 'mvdg 3/17/2007 22:20'!
xLetter
	"Form a word or keyword."

	| type c |
	buffer reset.
	[c _ hereChar asciiValue.
	((type _ typeTable at: c ifAbsent: [#xLetter]) == #xLetter or: [type == #xDigit])
		or: [(type == #period) and: [(typeTable at: (aheadChar asciiValue)) == #xLetter]]]
		whileTrue: ["open code step for speed"
			buffer nextPut: hereChar.
			hereChar _ aheadChar.
			source atEnd
				ifTrue: [aheadChar _ 30 asCharacter
					"doit"]
				ifFalse: [aheadChar _ source next]].
	(type == #colon or: [type == #xColon and: [aheadChar ~= $=]])
		ifTrue: [buffer nextPut: self step.
			["Allow any number of embedded colons in literal symbols"
			(typeTable at: hereChar asciiValue ifAbsent: [#xLetter])
				== #xColon]
				whileTrue: [buffer nextPut: self step].
			tokenType _ #keyword]
		ifFalse: [tokenType _ #word].
	token _ buffer contents.
	token isOctetString ifTrue: [token _ token asOctetString].
! !
